extern unsigned char mainram[];
extern unsigned char keym[];
extern unsigned char kon;
extern unsigned char disp_on, power_on;
extern unsigned char kbdpowoff;

/* main.c */
void kbdescape();
extern float sctiming;
void disprefresh();

/* semilcd.c */
void lcdinit();
void clearall();
void displayall();
void busy();

/* ttyin.c */
void kbdinit();
void kbdrestore();
int ttyin(int select_timeout);

/* sc.c */
void initsc();
void scstatus();
int interpret();
void closesc();
void dumpiram();
void dumpmainram(int start);
extern int key_wait;

/* key.c */
void showkeytab();
int keyscan(int ch);

